from functools import wraps


def authorization_level(package):
    """
    returns write, read or None
    """
    if package == "ant31/no":
        return None
    elif package == "w":
        return "write"
    else:
        return 'write'


def write_access(package):
    return (authorization_level(package) == 'write')


def read_access(package):
    return (authorization_level(package) is not None)


def check_access(mode):
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            package = kwargs['package']
            if mode == "write":
                access = write_access(package)
            elif mode == "read":
                access = read_access(package)
            else:
                raise ValueError("unknow access mode: %s" % mode)
            if not access:
                raise RuntimeError("Non authorized to access: %s" % package)
            else:
                return function(*args, **kwargs)
        return wrapper
    return decorator


@check_access('read')
def read(package):
    print "ok"


@check_access('write')
def write(package):
    print "ok"


@check_access('bad')
def bad(package):
    print "bad"



from flask import jsonify, Blueprint, current_app
import kpm.api.impl.builder
from kpm.api.app import getvalues

builder_app = Blueprint('builder', __name__,)


def _build(package):
    values = getvalues()
    version = values.get('version', None)
    namespace = values.get('namespace', 'default')
    variables = values.get('variables', {})
    shards = values.get('shards', None)
    variables['namespace'] = namespace
    k = kpm.api.impl.builder.build(package,
                                   version_query=version,
                                   namespace=namespace,
                                   variables=variables,
                                   shards=shards,
                                   endpoint=current_app.config['KPM_REGISTRY_HOST'])
    return k


@builder_app.route("/api/v1/packages/<path:package>/file/<path:filepath>")
def show_file(package, filepath):
    return kpm.api.impl.builder.show_file(package, filepath,
                                          endpoint=current_app.config['KPM_REGISTRY_HOST'])


@builder_app.route("/api/v1/packages/<path:package>/tree")
def tree(package):
    r = kpm.api.impl.builder.tree(package,
                                  endpoint=current_app.config['KPM_REGISTRY_HOST'])
    return jsonify(r)


@builder_app.route("/api/v1/packages/<path:package>/generate", methods=['POST', 'GET'])
def build(package):
    current_app.logger.info("generate %s", package)
    k = _build(package)
    return jsonify(k.build())


@builder_app.route("/api/v1/packages/<path:package>/generate-tar", methods=['POST', 'GET'])
def build_tar(package):
    k = _build(package)
    resp = current_app.make_response(k.build_tar())
    resp.mimetype = 'application/tar'
    resp.headers['Content-Disposition'] = 'filename="%s_%s.tar.gz"' % (k.name.replace("/", "_"), k.version)
    return resp



class CommandBase(object):
    name = 'command-base'
    help_message = 'describe the command'

    def __init__(self, args_options):
        pass

    def render(self):
        if self.output == 'json':
            self._render_json()
        else:
            self._render_console()

    @classmethod
    def call(self, options):
        self(options)()

    def __call__(self):
        self._call()
        self.render()

    @classmethod
    def add_parser(self, subparsers):
        parser = subparsers.add_parser(self.name, help=self.help_message)
        parser.add_argument("--output", default="text",  choices=['text', 'json'],
                            help="output format")
        self._add_arguments(parser)
        parser.set_defaults(func=self.call)

    def _render_json(self):
        raise NotImplementedError

    def _render_console(self):
        raise NotImplementedError

    def _call(self):
        raise NotImplementedError

    @classmethod
    def _add_arguments(self, parser):
        raise NotImplementedError



import json
from base64 import b64decode
from flask import jsonify, request, Blueprint, current_app
from kpm.api.app import getvalues
import kpm.api.impl.registry
from kpm.exception import (KpmException,
                           InvalidUsage,
                           InvalidVersion,
                           PackageAlreadyExists,
                           ChannelAlreadyExists,
                           PackageNotFound,
                           ChannelNotFound,
                           PackageVersionNotFound)

import etcd

registry_app = Blueprint('registry', __name__,)


@registry_app.errorhandler(etcd.EtcdKeyNotFound)
def render_etcdkeyerror(error):
    package = error.payload['cause']
    return render_error(PackageNotFound("Package not found: %s" % package, {"package": package}))


@registry_app.errorhandler(PackageAlreadyExists)
@registry_app.errorhandler(ChannelAlreadyExists)
@registry_app.errorhandler(InvalidVersion)
@registry_app.errorhandler(PackageNotFound)
@registry_app.errorhandler(PackageVersionNotFound)
@registry_app.errorhandler(KpmException)
@registry_app.errorhandler(InvalidUsage)
@registry_app.errorhandler(ChannelNotFound)
def render_error(error):
    response = jsonify({"error": error.to_dict()})
    response.status_code = error.status_code
    return response


@registry_app.route("/test_error")
def test_error():
    raise InvalidUsage("error message", {"path": request.path})


@registry_app.route("/api/v1/packages/<path:package>/pull", methods=['GET'], strict_slashes=False)
def pull(package):
    current_app.logger.info("pull %s", package)
    values = getvalues()
    version = values.get("version", 'latest')
    r = kpm.api.impl.registry.pull(package, version)
    if 'format' in values and values['format'] == 'json':
        resp = jsonify({"package": r['package'], "kub": r['blob']})
    else:
        resp = current_app.make_response(b64decode(r['blob']))
        resp.headers['Content-Disposition'] = r['filename']
        resp.mimetype = 'application/x-gzip'
    return resp


@registry_app.route("/api/v1/packages/<path:package>", methods=['POST'], strict_slashes=False)
@registry_app.route("/api/v1/packages", methods=['POST'], strict_slashes=False)
def push(package=None):
    values = getvalues()
    blob = values['blob']
    package = values['package']
    version = values['version']
    force = False
    if 'force' in values:
        force = 'true' == values['force']

    r = kpm.api.impl.registry.push(package, version, blob, force)
    return jsonify(r)


@registry_app.route("/api/v1/packages", methods=['GET'], strict_slashes=False)
def list_packages():
    values = getvalues()
    organization = values.get('organization', None)
    r = kpm.api.impl.registry.list_packages(organization=organization)
    resp = current_app.make_response(json.dumps(r))
    resp.mimetype = 'application/json'
    return resp


@registry_app.route("/api/v1/packages/<path:package>", methods=['GET'], strict_slashes=False)
def show_package(package):
    values = getvalues()
    version = values.get("version", 'latest')
    pullmode = False
    if 'pull' in values and (values['pull'] == 'true' or values['pull']):
        pullmode = True
    r = kpm.api.impl.registry.show_package(package, version, pullmode)
    return jsonify(r)


# CHANNELS
@registry_app.route("/api/v1/packages/<path:package>/channels", methods=['GET'], strict_slashes=False)
def list_channels(package):
    r = kpm.api.impl.registry.list_channels(package)
    resp = current_app.make_response(json.dumps(r))
    resp.mimetype = 'application/json'
    return resp


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>", methods=['GET'], strict_slashes=False)
def show_channel(package, name):
    r = kpm.api.impl.registry.show_channel(package, name)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>/<string:release>",
                    methods=['POST'], strict_slashes=False)
def add_channel_release(package, name, release):
    r = kpm.api.impl.registry.add_channel_release(package, name, release)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>/<string:release>",
                    methods=['DELETE'], strict_slashes=False)
def delete_channel_release(package, name, release):
    r = kpm.api.impl.registry.delete_channel_release(package, name, release)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>",
                    methods=['POST'], strict_slashes=False)
def create_channel(package, name):
    r = kpm.api.impl.registry.create_channel(package, name)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>",
                    methods=['DELETE'], strict_slashes=False)
def delete_channel(package, name):
    r = kpm.api.impl.registry.delete_channel(package, name)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<string:orga>/<string:pname>", methods=['DELETE'], strict_slashes=False)
def delete_package(orga, pname):
    package = "%s/%s" % (orga, pname)
    values = getvalues()
    version = values.get("version", "latest")
    r = kpm.api.impl.registry.delete_package(package, version)
    return jsonify(r)



#!/usr/bin/env python
import argparse
import os
import yaml
import json
import copy
import re
from kpm.render_jsonnet import RenderJsonnet
from kpm.commands import all_commands


class LoadVariables(argparse.Action):
    def _parse_cmd(self, var):
        r = {}
        try:
            return json.loads(var)
        except:
            for v in var.split(","):
                sp = re.match("(.+?)=(.+)", v)
                if sp is None:
                    raise ValueError("Malformed variable: %s" % v)
                key, value = sp.group(1), sp.group(2)
                r[key] = value
        return r

    def _load_from_file(self, filename, ext):
        with open(filename, 'r') as f:
            if ext in ['.yml', '.yaml']:
                return yaml.load(f.read())
            elif ext == '.json':
                return json.loads(f.read())
            elif ext in [".jsonnet", "libjsonnet"]:
                r = RenderJsonnet()
                return r.render_jsonnet(f.read())
            else:
                raise ValueError("File extension is not in [yaml, json, jsonnet], %s" % filename)

    def load_variables(self, var):
        _, ext = os.path.splitext(var)
        if ext == '':
            return self._parse_cmd(var)
        else:
            return self._load_from_file(var, ext)

    def __call__(self, parser, namespace, values, option_string=None):
        items = copy.copy(argparse._ensure_value(namespace, self.dest, {}))
        items.update(self.load_variables(values))
        setattr(namespace, self.dest, items)


def get_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='command help')
    for _, commandClass in all_commands.iteritems():
        commandClass.add_parser(subparsers)
    return parser


def cli():
    parser = get_parser()
    args = parser.parse_args()
    try:
        args.func(args)
    except (argparse.ArgumentTypeError, argparse.ArgumentError) as e:
        parser.error(e.message)



from kpm.commands.push import PushCmd
from kpm.commands.pull import PullCmd
from kpm.commands.version import VersionCmd
from kpm.commands.new import NewCmd
from kpm.commands.show import ShowCmd
from kpm.commands.login import LoginCmd
from kpm.commands.logout import LogoutCmd
from kpm.commands.deploy import DeployCmd
from kpm.commands.channel import ChannelCmd
from kpm.commands.list_package import ListPackageCmd
from kpm.commands.remove import RemoveCmd
from kpm.commands.delete_package import DeletePackageCmd
from kpm.commands.kexec import ExecCmd
from kpm.commands.generate import GenerateCmd
from kpm.commands.jsonnet import JsonnetCmd


all_commands = {
    PushCmd.name: PushCmd,
    VersionCmd.name: VersionCmd,
    PullCmd.name: PullCmd,
    NewCmd.name: NewCmd,
    ShowCmd.name: ShowCmd,
    LoginCmd.name: LoginCmd,
    LogoutCmd.name: LogoutCmd,
    DeployCmd.name: DeployCmd,
    ChannelCmd.name: ChannelCmd,
    DeletePackageCmd.name: DeletePackageCmd,
    ListPackageCmd.name: ListPackageCmd,
    RemoveCmd.name: RemoveCmd,
    ExecCmd.name: ExecCmd,
    JsonnetCmd.name: JsonnetCmd,
    GenerateCmd.name: GenerateCmd,

}



import os
import json
import base64
import kpm.registry
from kpm.manifest_jsonnet import ManifestJsonnet
from kpm.packager import pack_kub
from kpm.commands.command_base import CommandBase


class PushCmd(CommandBase):
    name = 'push'
    help_message = "push a package to the registry"

    def __init__(self, options):
        self.output = options.output
        self.registry_host = options.registry_host
        self.force = options.force
        self.manifest = None

        super(PushCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')
        parser.add_argument("-o", "--organization", nargs="?", default=None,
                            help="push to another organization")
        parser.add_argument("-f", "--force", action='store_true', default=False,
                            help="force push")

    def _call(self):
        r = kpm.registry.Registry(self.registry_host)
        # @TODO: Override organization
        self.manifest = ManifestJsonnet()
        # @TODO: Pack in memory
        kubepath = os.path.join(".", self.manifest.package_name() + "kub.tar.gz")
        pack_kub(kubepath)
        f = open(kubepath, 'rb')
        r.push(self.manifest.package['name'], {"name": self.manifest.package['name'],
                                               "version": self.manifest.package['version'],
                                               "blob": base64.b64encode(f.read())}, self.force)
        f.close()
        os.remove(kubepath)

    def _render_json(self):
        print json.dumps({"package": self.manifest.package['name'],
                          "version": self.manifest.package['version']})

    def _render_console(self):
        print "package: %s (%s) pushed" % (self.manifest.package['name'],
                                           self.manifest.package['version'])



import pytest
import kpm.utils
import os.path


def test_mkdirp_on_existing_dir(tmpdir):
    exists = str(tmpdir.mkdir("dir1"))
    kpm.utils.mkdir_p(exists)
    assert os.path.exists(exists)


def test_mkdirp(tmpdir):
    path = os.path.join(str(tmpdir), "new/directory/tocreate")
    kpm.utils.mkdir_p(path)
    assert os.path.exists(path)


def test_mkdirp_unauthorized(tmpdir):
    import os
    d = str(tmpdir.mkdir("dir2"))
    path = os.path.join(d, "directory/tocreate")
    os.chmod(d, 0)
    with pytest.raises(OSError):
        kpm.utils.mkdir_p(path)


def test_colorize():
    assert kpm.utils.colorize('ok') == "\x1b[32mok\x1b[0m"


# def test_parse_cmdline_variables():
#     l = ["titi=tata"]
#     assert {"titi": "tata"} == kpm.utils.parse_cmdline_variables(l)


# def test_parse_cmdline_variables_comma():
#     l = ["titi=tata,lola=popa"]
#     assert {"titi": "tata", "lola": "popa"} == kpm.utils.parse_cmdline_variables(l)


# def test_parse_cmdline_variables_multi():
#     l = ["titi=tata,lola=popa", "mami=papi"]
#     assert {"titi": "tata",
#             "lola": "popa",
#             "mami": "papi"} == kpm.utils.parse_cmdline_variables(l)


# def test_parse_cmdline_variables_multi_overwrite():
#     l = ["titi=tata,lola=popa", "titi=papi"]
#     assert {"titi": "papi", "lola": "popa"} == kpm.utils.parse_cmdline_variables(l)


# def test_parse_cmdline_variables_bad():
#     l = ["titi=tata,lola=popa", "titipapi"]
#     with pytest.raises(ValueError):
#         assert {"titi": "papi", "lola": "popa"} == kpm.utils.parse_cmdline_variables(l)


# def test_parse_cmdline_variables_json():
#     l = ['{"titi": ["tata", "lola"]}']
#     assert {"titi": ["tata", "lola"]} == kpm.utils.parse_cmdline_variables(l)


# def test_parse_cmdline_variables_mixjson():
#     l = ['{"titi": ["tata", "lola"]}', "test=test2"]
#     assert {"titi": ["tata", "lola"], "test": "test2"} == kpm.utils.parse_cmdline_variables(l)


def test_convert_utf8():
    data = {u"test": {u"test2": u"test3"},
            u"array": [u"a1", u"a2"], u"int": 5}
    assert {"test": {"test2": "test3"},
            "array": ["a1", "a2"], "int": 5} == kpm.utils.convert_utf8(data)



from urlparse import urlparse
import copy
import hashlib
import tarfile
import shutil
import logging
import os.path
import yaml
import io
import tempfile
import json
import kpm.registry as registry
import kpm.packager as packager
from kpm.utils import mkdir_p


logger = logging.getLogger(__name__)


_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


class KubBase(object):
    def __init__(self, name,
                 version=None,
                 variables=None,
                 shards=None,
                 namespace=None,
                 endpoint=None,
                 resources=None):
        if variables is None:
            variables = {}
        self.endpoint = endpoint

        self._dependencies = None
        self._resources = None
        self._deploy_name = name
        self._deploy_version = version
        self._deploy_shards = shards
        self._deploy_resources = resources
        self.namespace = namespace
        result = self._fetch_package()
        self.package = packager.Package(result, b64_encoded=False)
        if self.namespace:
            variables["namespace"] = self.namespace
        self.manifest = None
        self._deploy_vars = variables
        self._variables = None

    def __unicode__(self):
        return ("(<{class_name}({name}=={version})>".format(class_name=self.__class__.__name__,
                                                            name=self.name, version=self.version))

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __repr__(self):
        return unicode(self).encode('utf-8')

    @property
    def author(self):
        return self.manifest.package['author']

    @property
    def version(self):
        return self.manifest.package['version']

    @property
    def description(self):
        return self.manifest.package['description']

    @property
    def name(self):
        return self.manifest.package['name']

    @property
    def deploy(self):
        return self.manifest.deploy

    @property
    def variables(self):
        if self._variables is None:
            self._variables = copy.deepcopy(self.manifest.variables)
            self._variables.update(self._deploy_vars)
        return self._variables

    def _fetch_package(self):
        parse = urlparse(self._deploy_name)
        if parse.scheme in ["http", "https"]:
            # @TODO
            pass
        elif parse.scheme == "file":
            parts = parse.path.split("/")
            _, ext = os.path.splitext(parts[-1])
            if ext == "tar.gz":
                filepath = parse.path
            else:
                filepath = tempfile.NamedTemporaryFile().name
                packager.pack_kub(filepath)
            with open(filepath, "rb") as f:
                return f.read()

        elif parse.scheme == "":
            self._registry = registry.Registry(endpoint=self.endpoint)
            return self._registry.pull(self._deploy_name, self._deploy_version)
        else:
            raise ValueError("Unknown package (%s) scheme: %s" % (self._deploy_name, parse.scheme))

    @property
    def kubClass(self):
        raise NotImplementedError

    def _fetch_deps(self):
        self._dependencies = []
        for dep in self.manifest.deploy:
            if dep['name'] != '$self':
                variables = dep.get('variables', {})
                variables['kpmparent'] = {'name': self.name,
                                          'shards': self.shards,
                                          'variables': self.variables}

                kub = self.kubClass(dep['name'],
                                    endpoint=self.endpoint,
                                    version=dep.get('version', None),
                                    variables=variables,
                                    resources=dep.get('resources', None),
                                    shards=dep.get('shards', None),
                                    namespace=self.namespace)
                self._dependencies.append(kub)
            else:
                self._dependencies.append(self)

    def create_namespace(self, namespace):
        value = {"apiVersion": "v1",
                 "kind": "Namespace",
                 "metadata": {"name": namespace}}

        resource = {"file": "%s-ns.yaml" % namespace,
                    "name": namespace,
                    "generated": True,
                    "order": -1,
                    "hash": False,
                    "protected": True,
                    "value": value,
                    "patch": [],
                    "variables": {},
                    "type": "namespace"}
        return resource

    @property
    def dependencies(self):
        if self._dependencies is None:
            self._fetch_deps()
        return self._dependencies

    @property
    def shards(self):
        shards = self.manifest.shards
        if self._deploy_shards is not None and len(self._deploy_shards):
            shards = self._deploy_shards
        return shards

    def make_tarfile(self, source_dir):
        output = io.BytesIO()
        with tarfile.open(fileobj=output, mode="w:gz") as tar:
            tar.add(source_dir, arcname=os.path.basename(source_dir))
        return output

    def build(self):
        raise NotImplementedError

    def build_tar(self, dest="/tmp"):
        package_json = self.build()

        tempdir = tempfile.mkdtemp()
        dest = os.path.join(tempdir, self.manifest.package_name())
        mkdir_p(dest)
        index = 0
        for kub in self.dependencies:
            index = kub.prepare_resources(dest, index)

        with open(os.path.join(dest, ".package.json"), mode="w") as f:
            f.write(json.dumps(package_json))

        tar = self.make_tarfile(dest)
        tar.flush()
        tar.seek(0)
        shutil.rmtree(tempdir)
        return tar.read()

    # @TODO do it in jsonnet
    def _annotate_resource(self, kub, resource):
        sha = None
        if 'annotations' not in resource['value']['metadata']:
            resource['value']['metadata']['annotations'] = {}
        if resource.get('hash', True):
            sha = hashlib.sha256(json.dumps(resource['value'])).hexdigest()
            resource['value']['metadata']['annotations']['kpm.hash'] = sha
        resource['value']['metadata']['annotations']['kpm.version'] = kub.version
        resource['value']['metadata']['annotations']['kpm.package'] = kub.name
        resource['value']['metadata']['annotations']['kpm.parent'] = self.name
        resource['value']['metadata']['annotations']['kpm.protected'] = str(resource['protected']).lower()
        return resource



import json
import kpm.registry
import kpm.command
from kpm.kub_jsonnet import KubJsonnet
from kpm.commands.command_base import CommandBase


class GenerateCmd(CommandBase):
    name = 'generate'
    help_message = "Generate a package json"

    def __init__(self, options):
        self.output = options.output
        self.package = options.pull[0]
        self.version = options.version
        self.namespace = options.namespace
        self.variables = options.variables
        self.registry_host = options.registry_host
        self.kub = None
        super(GenerateCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument("--namespace", nargs="?",
                            help="kubernetes namespace", default='default')
        parser.add_argument("-x", "--variables",
                            help="variables", default={}, action=kpm.command.LoadVariables)
        parser.add_argument('-p', "--pull", nargs=1, help="Fetch package from the registry")
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')
        parser.add_argument("-v", "--version", nargs="?", default=None,
                            help="package version")

    def _call(self):
        name = self.package
        version = self.version
        namespace = self.namespace
        k = KubJsonnet(name, endpoint=self.registry_host,
                       variables=self.variables,
                       namespace=namespace, version=version)
        filename = "%s_%s.tar.gz" % (k.name.replace("/", "_"), k.version)
        with open(filename, 'wb') as f:
            f.write(k.build_tar("."))
        self.kub = k

    def _render_json(self):
        print json.dumps(self.kub.manifest, indent=2, separators=(',', ': '))

    def _render_console(self):
        self._render_json()



import json
import kpm.registry
from kpm.commands.command_base import CommandBase
from kpm.auth import KpmAuth


class LogoutCmd(CommandBase):
    name = 'logout'
    help_message = "logout"

    def __init__(self, options):
        self.output = options.output
        self.registry_host = options.registry_host
        self.status = None
        super(LogoutCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')

    def _call(self):
        KpmAuth().delete_token()
        self.status = "Logout complete"

    def _render_json(self):
        print json.dumps({"status": self.status})

    def _render_console(self):
        print " >>> %s" % self.status



import json
import kpm.new
import argparse
from kpm.commands.command_base import CommandBase


class NewCmd(CommandBase):
    name = 'new'
    help_message = "initiate a new package"

    def __init__(self, options):
        self.output = options.output
        self.package = options.package[0]
        self.with_comments = options.with_comments
        self.directory = options.directory
        self.path = None
        super(NewCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument('package', nargs=1, help="package-name")
        parser.add_argument("--directory",  nargs="?", default=".",
                            help="destionation directory")
        parser.add_argument("--with-comments", action='store_true', default=False,
                            help="Add 'help' comments to manifest")

    def _call(self):
        try:
            self.path = kpm.new.new_package(self.package, self.directory, self.with_comments)
        except ValueError as e:
            argparse.ArgumentTypeError(e.message)

    def _render_json(self):
        print json.dumps({"new": self.package,
                          "path": self.path})

    def _render_console(self):
        print "New package created in %s" % self.path



import os


class Config(object):
    """ Default configuration """
    DEBUG = False
    KUBE_APIMASTER = os.getenv('KUBE_APIMASTER', 'http://localhost:8001')
    KPM_URI = os.getenv('KPM_URI', "http://localhost:5000")
    KPM_REGISTRY_HOST = os.getenv('KPM_REGISTRY_HOST', KPM_URI)
    KPM_BUILDER_HOST = os.getenv('KPM_BUILDER_HOST', KPM_URI)
    KPM_MODELS_MODULE = os.getenv('KPM_MODELS_MODULE', "kpm.models.etcd")
    KPM_MODELS = os.getenv('KPM_MODELS', '{"Package": "kpm.models.etcd.package:Package"}')
    KPM_API_BACKEND = 'true'
    KPM_API_BUILDER = 'true'
    KPM_API_REGISTRY = 'true'


class ProductionConfig(Config):
    """ Production configuration """
    KPM_URI = 'https://api.kpm.sh'
    KPM_BACKEND = 'false'


class DevelopmentConfig(Config):
    """ Development configuration """
    DEBUG = True
#    KPM_URI = 'https://api.kpm.sh'
    KPM_URI = os.getenv('KPM_URI', "http://localhost:5000")



import jinja2
import string
import yaml
import json
import random
import hashlib
from base64 import b64decode, b64encode


def get_hash(data, hashtype='sha1'):
    h = hashlib.new(hashtype)
    h.update(data)
    return h.hexdigest()


def rand_string(size=32, chars=(string.ascii_letters + string.digits), seed=None):
    if seed == "":
        seed = None
    random.seed(seed)
    size = int(size)
    return ''.join(random.choice(chars) for _ in range(size))


def rand_alphanum(size=32, seed=None):
    size = int(size)
    return rand_string(size=size, seed=seed)


def rand_alpha(size=32, seed=None):
    size = int(size)
    return rand_string(size=size, chars=string.ascii_letters, seed=seed)


def randint(size=32, seed=None):
    size = int(size)
    return rand_string(size=size, chars=string.digits, seed=seed)


def gen_private_ecdsa():
    from ecdsa import SigningKey
    sk = SigningKey.generate()
    return sk.to_pem()


def gen_private_rsa():
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.asymmetric import rsa

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )
    return pem


def gen_private_dsa():
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.asymmetric import dsa

    private_key = dsa.generate_private_key(
        key_size=1024,
        backend=default_backend()
    )
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )
    return pem


all_privates = {}


def gen_privatekey(keytype='rsa', key='', seed=None):
    if seed is None:
        seed = rand_alphanum(128)
    k = seed + key

    if k not in all_privates:
        all_privates[k] = {}

    if keytype == "ecdsa":
        if 'ecdsa' not in all_privates[k]:
            all_privates[k][keytype] = gen_private_ecdsa()

    elif keytype == "rsa":
        if 'rsa' not in all_privates[k]:
            all_privates[k][keytype] = gen_private_rsa()

    elif keytype == "dsa":
        if 'dsa' not in all_privates[k]:
            all_privates[k][keytype] = gen_private_dsa()
    else:
        raise ValueError("Unknow private key type: %s" % keytype)

    return all_privates[k][keytype]


def jinja_env():
    from kpm.template_filters import jinja_filters
    jinjaenv = jinja2.Environment()
    jinjaenv.filters.update(jinja_filters())
    return jinjaenv


def jinja_template(val, env=None):
    from kpm.utils import convert_utf8
    jinjaenv = jinja_env()
    template = jinjaenv.from_string(val)
    if env is not None:
        variables = convert_utf8(json.loads(env))
    return template.render(variables)


def jsonnet(val, env=None):
    from kpm.render_jsonnet import RenderJsonnet
    from kpm.utils import convert_utf8
    r = RenderJsonnet()
    if env is not None:
        variables = convert_utf8(json.loads(env))
    return r.render_jsonnet(val, tla_codes=variables)


def json_to_yaml(value):
    """
    Serializes an object as YAML. Optionally given keyword arguments
    are passed to yaml.dumps(), ensure_ascii however defaults to False.
    """
    return yaml.safe_dump(json.loads(value))


def json_dumps(value, **kwargs):
    """
    Serializes an object as JSON. Optionally given keyword arguments
    are passed to json.dumps(), ensure_ascii however defaults to False.
    """
    kwargs.setdefault('ensure_ascii', False)
    return json.dumps(value, **kwargs)


def yaml_dumps(value):
    """
    Serializes an object as YAML. Optionally given keyword arguments
    are passed to yaml.dumps(), ensure_ascii however defaults to False.
    """
    return yaml.dump(value,  default_flow_style=True)


def json_loads(value):
    """
    Serializes an object as JSON. Optionally given keyword arguments
    are passed to json.dumps(), ensure_ascii however defaults to False.
    """
    return json.loads(value)


def yaml_loads(value):
    """
    Serializes an object as JSON. Optionally given keyword arguments
    are passed to json.dumps(), ensure_ascii however defaults to False.
    """
    return yaml.load(value)


def obj_loads(value):
    try:
        return json.loads(value)
    except ValueError:
        return yaml.load(value)


def jinja_filters():
    filters = {
        'json': json_dumps,
        'yaml': yaml_dumps,
        'get_hash': get_hash,
        'b64decode': b64decode,
        'b64encode': b64encode,
        'gen_privatekey': gen_privatekey,
        'rand_alphanum': rand_alphanum,
        'rand_alpha': rand_alpha
        }
    return filters


def jsonnet_callbacks():
    filters = {
        'hash': (('data', 'hashtype'), get_hash),
        'to_yaml': (('value',), json_to_yaml),
        'rand_alphanum': (('size', 'seed'), rand_alphanum),
        'rand_alpha': (('size', 'seed'), rand_alpha),
        'randint': (('size', 'seed'), randint),
        'jinja2': (('template', 'env'), jinja_template),
        'jsonnet': (('template', 'env'), jsonnet),
        'json_loads': (('jsonstr',), json_loads),
        'yaml_loads': (('jsonstr',), yaml_loads),
        'obj_loads': (('jsonstr',), obj_loads),
        'privatekey': (('keytype', "key", "seed"), gen_privatekey),
    }
    return filters



from flask import jsonify, request, Blueprint, current_app
import etcd
import kpm.deploy
from kpm.exception import (KpmException,
                           InvalidUsage,
                           InvalidVersion,
                           PackageAlreadyExists,
                           PackageNotFound,
                           PackageVersionNotFound)


deployment_app = Blueprint('deployment', __name__,)
etcd_client = etcd.Client(port=2379)

ETCD_PREFIX = "kpm/deployments/"


@deployment_app.errorhandler(PackageAlreadyExists)
@deployment_app.errorhandler(PackageNotFound)
@deployment_app.errorhandler(PackageVersionNotFound)
@deployment_app.errorhandler(KpmException)
@deployment_app.errorhandler(InvalidVersion)
@deployment_app.errorhandler(InvalidUsage)
def render_error(error):
    response = jsonify({"error": error.to_dict()})
    response.status_code = error.status_code
    return response


def _cmd(cmd, package):
    jsonbody = request.get_json(force=True, silent=True)
    values = request.values.to_dict()
    if jsonbody:
        values.update(jsonbody)
    params = {"version": values.get("version"),
              "namespace": values.get("namespace"),
              "dry": values.get("dry", False) == 'true',
              "variables": values.get("variables", None),
              "endpoint": current_app.config['KPM_REGISTRY_HOST'],
              "proxy": current_app.config['KUBE_APIMASTER'],
              "fmt": "json"}
    current_app.logger.info("%s %s: %s", cmd, package, params)
    return getattr(kpm.deploy, cmd)(package, **params)


@deployment_app.route("/api/v1/deployments/<path:package>", methods=['DELETE'], strict_slashes=False)
def remove(package):
    r = _cmd('delete', package)
    return jsonify({"result": r})


@deployment_app.route("/api/v1/deployments/<path:package>", methods=['POST'], strict_slashes=False)
def deploy(package):
    r = _cmd('deploy', package)
    return jsonify({"result": r})



import json
import kpm.registry
from kpm.commands.command_base import CommandBase


class DeletePackageCmd(CommandBase):
    name = 'delete-package'
    help_message = 'delete package from the registry'

    def __init__(self, options):
        self.output = options.output
        self.package = options.package[0]
        self.registry_host = options.registry_host
        self.version = options.version
        self.result = None
        super(DeletePackageCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument('package', nargs=1, help="package-name")
        parser.add_argument("-v", "--version", nargs="?",
                            help="package VERSION", default=None)
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')

    def _call(self):
        r = kpm.registry.Registry(self.registry_host)
        self.result = r.delete_package(self.package, version=self.version)

    def _render_json(self):
        print json.dumps(self.result)

    def _render_console(self):
        print "Deleted package: %s - %s" % (self.result['package'], self.result['version'])



import json
import kpm.command
from kpm.render_jsonnet import RenderJsonnet
from kpm.commands.command_base import CommandBase


class JsonnetCmd(CommandBase):
    name = 'jsonnet'
    help_message = "Resolve a jsonnet file with the kpmstd available"

    def __init__(self, options):
        self.output = options.output
        self.shards = options.shards
        self.namespace = options.namespace
        self.variables = options.variables
        self.filepath = options.filepath[0]
        self.result = None

        super(JsonnetCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument("--namespace", nargs="?",
                            help="kubernetes namespace", default='default')
        parser.add_argument("-x", "--variables",
                            help="variables", default={}, action=kpm.command.LoadVariables)
        # @TODO shards
        parser.add_argument("--shards",
                            help="Shards list/dict/count: eg. --shards=5 ; --shards='[{\"name\": 1, \"name\": 2}]'",
                            default=None)
        parser.add_argument('filepath', nargs=1, help="Fetch package from the registry")

    def _call(self):
        r = RenderJsonnet()
        namespace = self.namespace
        self.variables['namespace'] = namespace
        tla_codes = {"variables": self.variables}
        p = open(self.filepath).read()
        self.result = r.render_jsonnet(p, tla_codes={"params": json.dumps(tla_codes)})

    def _render_json(self):
        print json.dumps(self.result, indent=2, separators=(',', ': '))

    def _render_console(self):
        self._render_json()



import logging

logger = logging.getLogger(__name__)


class ManifestBase(dict):
    def __init__(self):
        super(ManifestBase, self).__init__()

    @property
    def resources(self):
        return self.get("resources", [])

    @property
    def deploy(self):
        return self.get("deploy", [])

    @property
    def dependencies(self):
        return [x['name'] for x in self.deploy if x['name'] != "$self"]

    @property
    def variables(self):
        return self.get("variables", {})

    @property
    def package(self):
        return self.get("package", {})

    @property
    def shards(self):
        return self.get("shards", [])

    def kubname(self):
        sp = self.package['name'].split('/')
        name = "%s_%s" % (sp[0], sp[1])
        return name

    def package_name(self):
        package = ("%s_%s" % (self.kubname(), self.package['version']))
        return package

    def to_dict(self):
        return ({
            "package": self.package,
            "variables": self.variables,
            "resources": self.resources,
            "deploy": self.deploy})



import sys
import _jsonnet
import json
from kpm.template_filters import jsonnet_callbacks
from kpm.render_jsonnet import RenderJsonnet

#r = RenderJsonnet()
#result = r.render_jsonnet(open(sys.argv[1]).read())
def native_bool(b):
    return ['true', True, False, 1, 0]

json_str = _jsonnet.evaluate_file(
    sys.argv[1],
    native_callbacks={"nativeBool": (("bool",), native_bool)},
)

sys.stdout.write(json_str)
#sys.stdout.write(json.dumps(result))



from kpm.kub_jsonnet import KubJsonnet


DEFAULT_ENDPOINT = "http://localhost:5000"


def build(package,
          version_query=None,
          namespace="default",
          variables={},
          shards=None,
          endpoint=DEFAULT_ENDPOINT):
    """
    A build is the construction/expansion of a package.
    The result are the expanded/ready to use resources, including every dependencies.

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      version_query (:obj:`str`): a version query, eg: ">=1.5,<2.0"
      namespace (:obj:`str`): kubernetes namespace to deploy the resource
      variables (:obj:`dict`): override default package variables to resolve templated resources
      shards (:obj:`json`, :obj:`int`): the shards (experimental)
      endpoint (:obj:`str`): the kpm-registry server

    Returns:
      :obj:`kpm.kub_jsonnet.KubJsonnet`: the KubJsonnet object.
      To generate the build runs one of the following commands:
        * :obj:`kpm.kub_jsonnet.KubJsonnet.build()`: create a `dict`
        * :obj:`kpm.kub_jsonnet.KubBase.build_tar()`: create a `tar.gz`

    Example:
      Flask route example::

         @builder_app.route("/api/v1/packages/<path:package>/generate", methods=['GET'])
         def build(package):
            k = kpm.api.impl.builder.build("ant31/postgresql",
                                           version=">=9.5.0",
                                           namespace="db",
                                           variables={'memory': "8Gi"}
                                           shards=None,
                                           endpoint="http://localhost:5000")
            build = k.build()
            return jsonify(build)

    See Also:
       * :obj:`kpm.api.builder.build`
       * :obj:`kpm.api.builder.build_tar`

    """
    variables['namespace'] = namespace
    k = KubJsonnet(package,
                   endpoint=endpoint,
                   variables=variables,
                   namespace=namespace,
                   version=version_query,
                   shards=shards)
    return k


def show_file(package, filepath, version_query=None, endpoint=DEFAULT_ENDPOINT):
    """
    Returns the content of any file inside a package.
    Useful to navigate and inspect a package from a web-browser

    Args:
      package (:obj:`str`): package name in the format `namespace/name` or `domain.com/name`
      filepath (:obj:`str`): filepath relative to the package, eg: `templates/svc.yaml`
      version_query (:obj:`str`): a version query, eg: ">=1.5,<2.0"
      endpoint (:obj:`str`): the kpm-registry server

    Returns:
      :obj:`str`: the file content

    See Also:
       * :obj:`kpm.api.builder.show_file`

    """
    k = KubJsonnet(package, version=version_query, endpoint=endpoint)
    return k.package.file(filepath)


def tree(package, version_query=None, endpoint=DEFAULT_ENDPOINT):
    """
    List recursivly the files inside a package.

    Args:
      package (:obj:`str`): package name in the format `namespace/name` or `domain.com/name`
      version_query (:obj:`str`): a version query, eg: ">=1.5,<2.0"
      filepath (:obj:`str`): filepath relative to the package, eg: `templates/svc.yaml`
      endpoint (:obj:`str`): the kpm-registry server

    Returns:
      :obj:`list`: package file list::

    Example:
       >>> kpm.api.impl.builder.tree("ant31/rocketchat", "latest", "http://localhost:5000")
           [
            "README.md",
            "manifest.jsonnet",
            "manifest.yaml",
            "templates/rocketchat-rc.yml",
            "templates/rocketchat-svc.yml"
           ]


    See Also:
       * :obj:`kpm.api.builder.tree`

    """

    k = KubJsonnet(package, version=version_query, endpoint=endpoint)
    return k.package.tree()



import json
import kpm.registry
import kpm.packager
import kpm.manifest
import kpm.manifest_jsonnet
from kpm.display import print_packages
from kpm.commands.command_base import CommandBase


class ListPackageCmd(CommandBase):
    name = 'list'
    help_message = "list packages"

    def __init__(self, options):
        self.output = options.output
        self.registry_host = options.registry_host
        self.user = options.user
        self.organization = options.organization
        self.result = None
        super(ListPackageCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument("-u", "--user", nargs="?", default=None,
                            help="list packages owned by USER")
        parser.add_argument("-o", "--organization", nargs="?", default=None,
                            help="list ORGANIZATION packages")
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')

    def _call(self):
        r = kpm.registry.Registry(self.registry_host)
        self.result = r.list_packages(user=self.user, organization=self.organization)

    def _render_json(self):
        print json.dumps(self.result)

    def _render_console(self):
        print_packages(self.result)



import logging
import os.path
import yaml
import json
from collections import OrderedDict
import jsonpatch
import kpm.manifest_jsonnet as manifest
from kpm.kubernetes import get_endpoint
from kpm.utils import convert_utf8
from kpm.kub_base import KubBase


logger = logging.getLogger(__name__)


_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


class KubJsonnet(KubBase):
    def __init__(self, *args, **kwargs):
        shards = kwargs.get("shards", None)
        if shards.__class__ in [str, unicode]:
            shards = json.loads(shards)
            kwargs['shards'] = shards

        super(KubJsonnet, self).__init__(*args, **kwargs)

        self.tla_codes = {"variables": self._deploy_vars}
        if shards is not None:
            self.tla_codes["shards"] = shards

        self.manifest = manifest.ManifestJsonnet(self.package, {"params": json.dumps(self.tla_codes)})

    def _create_namespaces(self):
        if self.namespace:
            ns = self.create_namespace(self.namespace)
            self._resources.insert(0, ns)

    @property
    def kubClass(self):
        return KubJsonnet

    @property
    def dependencies(self):
        if self._dependencies is None:
            self._fetch_deps()
        return self._dependencies

    def _init_resources(self):
        index = 0
        for resource in self._resources:
            index += 1
            resource["order"] = index
            if 'protected' not in resource:
                resource["protected"] = False

    def resources(self):
        if self._resources is None:
            self._resources = self.manifest.resources
            self._create_namespaces()
        return self._resources

    def prepare_resources(self, dest="/tmp", index=0):
        for resource in self.resources():
            index += 1
            path = os.path.join(dest, "%02d_%s_%s" % (index,
                                                      self.version,
                                                      resource['file']))
            f = open(path, 'w')
            f.write(yaml.safe_dump(convert_utf8(resource['value'])))
            resource['filepath'] = f.name
            f.close()
        return index

    def build(self):
        result = []
        for kub in self.dependencies:
            kubresources = OrderedDict([("package",  kub.name),
                                        ("version", kub.version),
                                        ("namespace", kub.namespace),
                                        ("resources", [])])
            for resource in kub.resources():
                self._annotate_resource(kub, resource)
                kubresources['resources'].\
                    append(OrderedDict({"file": resource['file'],
                                        "hash": resource['value']['metadata']['annotations'].get('kpm.hash', None),
                                        "protected": resource['protected'],
                                        "name": resource['name'],
                                        "kind": resource['value']['kind'].lower(),
                                        "endpoint": get_endpoint(
                                            resource['value']['kind'].lower()).
                                        format(namespace=self.namespace),
                                        "body": json.dumps(resource['value'])}))

            result.append(kubresources)
        return {"deploy": result,
                "package": {"name": self.name,
                            "version": self.version}}

    def _apply_patches(self, resources):
        for _, resource in resources.iteritems():
            if self.namespace:
                if 'namespace' in resource['value']['metadata']:
                    op = 'replace'
                else:
                    op = 'add'
                resource['patch'].append({"op": op, "path": "/metadata/namespace", "value": self.namespace})

            if len(resource['patch']):
                patch = jsonpatch.JsonPatch(resource['patch'])
                result = patch.apply(resource['value'])
                resource['value'] = result
        return resources



import json
import kpm.registry
import kpm.packager
import kpm.manifest
import kpm.manifest_jsonnet
from kpm.commands.command_base import CommandBase


class ShowCmd(CommandBase):
    name = 'show'
    help_message = "print the package manifest"

    def __init__(self, options):
        self.output = options.output
        self.package = options.package[0]
        self.registry_host = options.registry_host
        self.version = options.version
        self.file = options.file
        self.tree = options.tree
        self.result = None
        super(ShowCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument('package', nargs=1, help="package-name")
        parser.add_argument('--tree', help="List files inside the package", action='store_true', default=False)
        parser.add_argument('-f', '--file', nargs="?", help="Display a file", default=None)
        parser.add_argument("-v", "--version", nargs="?", default=None,
                            help="package version")
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')

    def _call(self):
        r = kpm.registry.Registry(self.registry_host)
        result = r.pull(self.package, version=self.version)
        p = kpm.packager.Package(result, b64_encoded=False)
        if self.tree:
            self.result = "\n".join(p.tree())
        elif self.file:
            self.result = p.file(self.file)
        else:
            self.result = p.manifest

    def _render_json(self):
        print json.dumps({"show": self.package,
                          "output": self.result})

    def _render_console(self):
        print self.result



import os
from flask import Flask, request
from flask_cors import CORS
from kpm.loghandler import init_logging


def getvalues():
    jsonbody = request.get_json(force=True, silent=True)
    values = request.values.to_dict()
    if jsonbody:
        values.update(jsonbody)
    return values


def create_app():
    app = Flask(__name__)
    CORS(app)
    setting = os.getenv('APP_ENV', "development")

    if setting != 'production':
        app.config.from_object('kpm.api.config.DevelopmentConfig')
    else:
        app.config.from_object('kpm.api.config.ProductionConfig')
    from kpm.api.builder import builder_app
    from kpm.api.info import info_app
    from kpm.api.registry import registry_app
    from kpm.api.deployment import deployment_app

    if app.config['KPM_API_BUILDER'] == "true":
        app.register_blueprint(builder_app)
    app.register_blueprint(info_app)
    if app.config['KPM_API_REGISTRY'] == "true":
        app.register_blueprint(registry_app)
    if app.config['KPM_API_BACKEND'] == "true":
        app.register_blueprint(deployment_app)
    init_logging(app.logger, loglevel='INFO')
    app.logger.info("Start service")
    return app


if __name__ == "__main__":
    app = create_app()
    app.run(host='0.0.0.0')



import etcd
import re

ETCD_PREFIX = "kpm/packages/"


class EtcdClient(etcd.client.Client):
    pass


etcd_client = EtcdClient(port=2379)


def etcd_listkeys(key, path):
    result = []
    for child in key.children:
        m = re.match("^/%s/(.+)$" % path, child.key)
        if m is None:
            continue
        result.append(m.group(1))
    return result



from kpm.console import KubernetesExec
from kpm.commands.command_base import CommandBase


class ExecCmd(CommandBase):
    name = 'exec'
    help_message = "exec a command in pod from the RC or RS name.\
    It executes the command on the first matching pod'"

    def __init__(self, options):
        self.output = options.output
        self.kind = options.kind
        self.container = options.container
        self.namespace = options.namespace
        self.resource = options.name
        self.cmd = options.cmd
        self.result = None

        super(ExecCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument('cmd', nargs='+', help="command to execute")
        parser.add_argument("--namespace", nargs="?",
                            help="kubernetes namespace", default='default')

        parser.add_argument('-k', '--kind', choices=['deployment', 'rs', 'rc'], nargs="?",
                            help="deployment, rc or rs", default='rc')
        parser.add_argument('-n', '--name', help="resource name", default='rs')
        parser.add_argument('-c', '--container', nargs='?', help="container name", default=None)

    def _call(self):
        c = KubernetesExec(self.resource,
                           cmd=" ".join(self.cmd),
                           namespace=self.namespace,
                           container=self.container,
                           kind=self.kind)
        self.result = c.call()

    def _render_json(self):
        pass

    def _render_console(self):
        print self.result



import json
import argparse
import kpm.registry
from kpm.commands.command_base import CommandBase


class ChannelCmd(CommandBase):
    name = 'channel'
    help_message = "Manage package channels"

    def __init__(self, options):
        self.output = options.output
        self.package = options.package[0]
        self.registry_host = options.registry_host
        self.create = options.create
        self.delete = options.delete
        self.channel = options.name
        self.remove = options.remove_release
        self.add = options.add_release
        self.list = options.list
        self.status = None
        self.channels = {}
        super(ChannelCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument("-n", "--name", nargs="?", default=None,
                            help="channel name")
        parser.add_argument("--add-release", nargs="?", default=None,
                            help="Add a release to the channel")
        parser.add_argument("--create", default=False, action='store_true',
                            help="Create the channel")
        parser.add_argument("--list", default=False, action='store_true',
                            help="Create the channel")
        parser.add_argument("--delete", default=False, action='store_true',
                            help="delete the channel")
        parser.add_argument("--remove-release", nargs="?", default=None,
                            help="Remove a release from the channel")
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')
        parser.add_argument('package', nargs=1, help="package-name")

    def _call(self):
        r = kpm.registry.Registry(self.registry_host)
        package = self.package
        name = self.channel
        if name is None and not self.list:
            raise argparse.ArgumentError(self.channel, "missing channel name")

        if self.create is True:
            self.channels = r.create_channel(package, name)
            self.status = "Channel '%s' on '%s' created" % (name, package)
        elif self.delete is True:
            self.channels = r.delete_channel(package, name)
            self.status = "Channel '%s' on '%s' deleted" % (name, package)
        elif self.list:
            self.channels = r.show_channels(package, name)
            # @TODO
            self.status = self.channels
        else:
            if self.add is not None:
                self.channels = r.create_channel_release(package, name, self.add)
                self.status = ">>> Release '%s' added on '%s'" % (self.add, name)
            if self.remove is not None:
                self.channels = r.delete_channel_release(package, name, self.remove)
                self.status = ">>> Release '%s' removed from '%s'" % (self.remove, name)

    def _render_json(self):
        print json.dumps(self.channels)

    def _render_console(self):
        print " >>> %s" % self.status



import re
import etcd
from kpm.models.channel_base import ChannelBase
from kpm.models.etcd import etcd_listkeys, ETCD_PREFIX, etcd_client


class Channel(ChannelBase):
    def __init__(self, name, package):
        super(Channel, self).__init__(name, package)

    @classmethod
    def _etcdpath(self, package, name=None):
        path = ETCD_PREFIX + package + "/channels"
        if name is not None:
            path = path + "/%s" % name
        return path

    def _exists(self):
        """ Check if the channel is saved already """
        path = self._etcdpath(self.package, self.name)
        try:
            etcd_client.read(path)
        except etcd.EtcdKeyNotFound:
            return False
        return True

    def save(self):
        path = self._etcdpath(self.package, self.name)
        try:
            etcd_client.write(path, None,  dir=True, prevExist=False)
            return True
        except etcd.EtcdAlreadyExist:
            raise self._raise_already_exists(self.name, self.package)

    def delete(self):
        path = self._etcdpath(self.package, self.name)
        if self.exists:
            etcd_client.delete(path, recursive=True)
            return True
        return False

    @classmethod
    def all(self, package):
        """
        Returns all available channels for a package
        """
        path = self._etcdpath(package)
        try:
            chans = etcd_client.read(path, recursive=True)
        except etcd.EtcdKeyNotFound:
            return []
        result = []
        for child in chans.children:
            m = re.match("^/%s/([a-zA-Z0-9-_]+)/?.*$" % path, child.key)
            if m is None:
                continue
            channel = m.group(1)
            result.append(channel)
        return result

    def releases(self):
        """ Returns the list of versions """
        path = self._etcdpath(self.package, self.name)
        try:
            releases = etcd_client.read(path, recursive=True)
        except etcd.EtcdKeyNotFound:
            self._raise_not_found(self.package, self.name)
        result = etcd_listkeys(releases, path)
        return result

    def _add_release(self, version):
        path = self._etcdpath(self.package, "%s/%s" % (self.name, version))
        try:
            n = self._new_chan_release(version)
            etcd_client.write(path, n, prevExist=False)
        except etcd.EtcdAlreadyExist:
            raise self._raise_already_exists(self.name, self.package, version)
        return n

    def _remove_release(self, version):
        path = self._etcdpath(self.package, "%s/%s" % (self.name, version))
        try:
            etcd_client.delete(path)
            return True
        except etcd.EtcdKeyNotFound:
            self._raise_not_found(self.package, self.name, version)

    def activate_release(self, version):
        raise NotImplementedError



import pytest
from kpm.manifest_jsonnet import ManifestJsonnet


@pytest.fixture()
def manifest(kubeui_package, package_dir):
    return ManifestJsonnet(kubeui_package)

@pytest.fixture()
def empty_manifest(empty_package_dir):
    return ManifestJsonnet(package=None)

@pytest.fixture()
def bad_manifest():
    return ManifestJsonnet(package=None)


def test_empty_resources(empty_manifest):
    assert empty_manifest.resources == []


def test_empty_variables(empty_manifest):
    assert empty_manifest.variables == {'namespace': 'default'}


def test_empty_package(empty_manifest):
    assert empty_manifest.package == {'expander': "jinja2"}


def test_empty_shards(empty_manifest):
    assert empty_manifest.shards is None


def test_empty_deploy(empty_manifest):
    assert empty_manifest.deploy == []


def test_package_name(manifest):
    assert manifest.package_name() == "kube-system_kube-ui_1.0.1"


def test_kubename(manifest):
    assert manifest.kubname() == "kube-system_kube-ui"


def test_load_from_path(manifest):
    m = ManifestJsonnet()
    assert m == manifest


def test_load_bad_manifest(bad_package_dir):
    import yaml
    with pytest.raises(yaml.YAMLError):
        ManifestJsonnet(package=None)



from __future__ import absolute_import
import importlib
import sys
import errno
import os
import collections
from termcolor import colored


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def colorize(status):
    msg = {'ok': 'green',
           'created': 'yellow',
           'updated': 'yellow',
           'absent': 'green',
           'deleted': 'red',
           'protected': 'blue'}
    return colored(status, msg[status])


def convert_utf8(data):
    if isinstance(data, basestring):
        return str(data)
    elif isinstance(data, collections.Mapping):
        return dict(map(convert_utf8, data.iteritems()))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(convert_utf8, data))
    else:
        return data


def custom_import(name):
    import importlib
    module, klass_name = name.split(':')
    mod = importlib.import_module(module)
    klass = getattr(mod, klass_name)
    return klass


# from celery/kombu https://github.com/celery/celery (BSD license)
def symbol_by_name(name, aliases={}, imp=None, package=None,
                   sep='.', default=None, **kwargs):
    """Get symbol by qualified name.

    The name should be the full dot-separated path to the class::

        modulename.ClassName

    Example::

        celery.concurrency.processes.TaskPool
                                    ^- class name

    or using ':' to separate module and symbol::

        celery.concurrency.processes:TaskPool

    If `aliases` is provided, a dict containing short name/long name
    mappings, the name is looked up in the aliases first.

    Examples:

        >>> symbol_by_name('celery.concurrency.processes.TaskPool')
        <class 'celery.concurrency.processes.TaskPool'>

        >>> symbol_by_name('default', {
        ...     'default': 'celery.concurrency.processes.TaskPool'})
        <class 'celery.concurrency.processes.TaskPool'>

        # Does not try to look up non-string names.
        >>> from celery.concurrency.processes import TaskPool
        >>> symbol_by_name(TaskPool) is TaskPool
        True

    """
    def _reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    if imp is None:
        imp = importlib.import_module

    if not isinstance(name, basestring):
        return name  # already a class

    name = aliases.get(name) or name
    sep = ':' if ':' in name else sep
    module_name, _, cls_name = name.rpartition(sep)
    if not module_name:
        cls_name, module_name = None, package if package else cls_name
    try:
        try:
            module = imp(module_name, package=package, **kwargs)
        except ValueError as exc:
            _reraise(ValueError,
                     ValueError("Couldn't import {0!r}: {1}".format(name, exc)),
                     sys.exc_info()[2])
        return getattr(module, cls_name) if cls_name else module
    except (ImportError, AttributeError):
        if default is None:
            raise
    return default



import os
import json
import kpm.registry
import kpm.packager
import kpm.command
from kpm.manifest_jsonnet import ManifestJsonnet
from kpm.commands.command_base import CommandBase


class PullCmd(CommandBase):
    name = 'pull'
    help_message = "download a package and extract it"

    def __init__(self, options):
        self.output = options.output
        self.package = options.package[0]
        self.registry_host = options.registry_host
        self.version = options.version
        self.dest = options.dest
        self.path = None
        super(PullCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument('package', nargs=1, help="package-name")
        parser.add_argument("--dest", nargs="?", default="/tmp/",
                            help="directory used to extract resources")
        parser.add_argument("-v", "--version", nargs="?",
                            help="package VERSION", default=None)
        parser.add_argument("-x", "--variables",
                            help="variables", default={}, action=kpm.command.LoadVariables)
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')

    def _call(self):
        r = kpm.registry.Registry(self.registry_host)
        result = r.pull(self.package, version=self.version)
        p = kpm.packager.Package(result, b64_encoded=False)
        self.path = os.path.join(self.dest, ManifestJsonnet(p).package_name())
        p.extract(self.path)

    def _render_json(self):
        print json.dumps({"pull": self.package,
                          "extrated": self.path})

    def _render_console(self):
        print "Pull package: %s... \nExtract to %s" % (self.package, self.path)



import json
import kpm.deploy
import kpm.command
from kpm.commands.command_base import CommandBase


class DeployCmd(CommandBase):
    name = 'deploy'
    help_message = "deploy a package on kubernetes"

    def __init__(self, options):
        self.output = options.output
        self.package = options.package[0]
        self.registry_host = options.registry_host
        self.shards = options.shards
        self.force = options.force
        self.dry_run = options.dry_run
        self.namespace = options.namespace
        self.api_proxy = options.api_proxy
        self.version = options.version
        self.tmpdir = options.tmpdir
        self.variables = options.variables
        self.status = None
        super(DeployCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument('package', nargs=1, help="package-name")
        parser.add_argument("--tmpdir", nargs="?", default="/tmp/",
                            help="directory used to extract resources")
        parser.add_argument("--dry-run", action='store_true', default=False,
                            help="do not create the resources on kubernetes")
        parser.add_argument("--namespace", nargs="?",
                            help="kubernetes namespace", default=None)
        parser.add_argument("--api-proxy", nargs="?",
                            help="kubectl proxy url", const="http://localhost:8001")
        parser.add_argument("-v", "--version", nargs="?",
                            help="package VERSION", default=None)
        parser.add_argument("-x", "--variables",
                            help="variables", default={}, action=kpm.command.LoadVariables)
        parser.add_argument("--shards",
                            help="Shards list/dict/count: eg. --shards=5 ; --shards='[{\"name\": 1, \"name\": 2}]'",
                            default=None)
        parser.add_argument("--force", action='store_true', default=False,
                            help="force upgrade, delete and recreate resources")
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')

    def _call(self):
        self.status = kpm.deploy.deploy(self.package,
                                        version=self.version,
                                        dest=self.tmpdir,
                                        namespace=self.namespace,
                                        force=self.force,
                                        dry=self.dry_run,
                                        endpoint=self.registry_host,
                                        proxy=self.api_proxy,
                                        variables=self.variables,
                                        shards=self.shards,
                                        fmt=self.output)

    def _render_json(self):
        print json.dumps(self.status)

    def _render_console(self):
        """ Handled by deploy """
        pass



# -*- coding: utf-8 -*-
import os
from kpm.utils import symbol_by_name

__author__ = 'Antoine Legrand'
__email__ = '2t.antoine@gmail.com'
__version__ = '0.21.1'


def version(registry_host=None):
    import requests
    from kpm.registry import Registry
    api_version = None
    ctl_version = __version__
    try:
        r = Registry(registry_host)
        response = r.version()
        api_version = response
    except requests.exceptions.RequestException:
        api_version = ".. Connection error"

    return {'api-version': api_version,
            "client-version": ctl_version}

models_path = {
    'Package': '{models_module}.package:Package',
    'Channel': '{models_module}.channel:Channel',
}

models_module = os.getenv('KPM_MODELS_MODULE', "kpm.models.etcd")


def setup_models(importlib, models=None, module=None):
    """Attach all model classes given by `models_path` to `importlib`."""
    if models is None:
        models = models_path
    if module is not None:
        models = {k: v.format(models_module=models_module) for k, v in models.iteritems()}
    importlib.__all__ = []
    for class_name, path in models.iteritems():
        orm_cls = symbol_by_name(path)
        setattr(importlib, class_name, orm_cls)
        importlib.__all__.append(class_name)



import kpm.deploy
from kpm.commands.deploy import DeployCmd


class RemoveCmd(DeployCmd):
    name = 'remove'
    help_message = "remove a package from kubernetes"

    def _call(self):
        self.status = kpm.deploy.delete(self.package,
                                        version=self.version,
                                        dest=self.tmpdir,
                                        namespace=self.namespace,
                                        force=self.force,
                                        dry=self.dry_run,
                                        endpoint=self.registry_host,
                                        proxy=self.api_proxy,
                                        variables=self.variables,
                                        shards=self.shards,
                                        fmt=self.output)



import yaml
import json
import _jsonnet
import re
import os
import logging
import os.path
import jinja2
from kpm.utils import convert_utf8
import kpm.template_filters as filters

logger = logging.getLogger(__name__)


with open(os.path.join(os.path.dirname(__file__), "jsonnet/manifest.jsonnet.j2")) as f:
    JSONNET_TEMPLATE = f.read()


def yaml_to_jsonnet(manifestyaml, tla_codes=None):
    jinja_env = jinja2.Environment()
    jinja_env.filters.update(filters.jinja_filters())
    # 1. Resolve old manifest variables
    # Load 'old' manifest.yaml
    v = {"manifest": convert_utf8(json.loads(json.dumps(yaml.load(manifestyaml))))}
    # Get variable from the 'old' manfiest and update  them
    variables = v['manifest'].get("variables", {})
    if tla_codes is not None and 'params' in tla_codes:
        tla = json.loads(tla_codes['params']).get("variables", {})
        variables.update(tla)
    # Resolve the templated variables inside the 'old' manifest
    manifest_tpl = jinja_env.from_string(manifestyaml)

    # 2. Convert 'old' manifest.yaml to manifest.jsonnet
    rendered_manifestyaml = manifest_tpl.render(variables)
    v = {"manifest": convert_utf8(json.loads(json.dumps(yaml.load(rendered_manifestyaml))))}
    # Load the yaml -> jsonnet template
    template = jinja_env.from_string(JSONNET_TEMPLATE)
    templatedjsonnet = template.render(v)
    # @TODO keep yaml format and escape 'jsonnet' commands:  key: "<% $.variables.key %>"
    # jsonnet_str = re.sub(r'[\'"]<%(.*)%>["\']', r"\1", templatedjsonnet)
    return templatedjsonnet


class RenderJsonnet(object):
    def __init__(self, files=None):
        self.files = files

    #  Returns content if worked, None if file not found, or throws an exception
    def try_path(self, path, rel):
        if not rel:
            raise RuntimeError('Got invalid filename (empty string).')
        if self.files is not None and rel in self.files:
            if self.files[rel] is None:
                with open(rel) as f:
                    self.files[rel] = f.read()
            return rel, self.files[rel]

        if rel in ["kpm.libjsonnet", "kpm-utils.libjsonnet"]:
            with open(os.path.join(os.path.dirname(__file__), "jsonnet/lib/%s" % rel)) as f:
                return rel, f.read()

        if rel[0] == '/':
            full_path = rel
        else:
            full_path = path + rel
        if full_path[-1] == '/':
            raise RuntimeError('Attempted to import a directory')
        if not os.path.isfile(full_path):
            return full_path, None
        with open(full_path) as f:
            return full_path, f.read()

    def import_callback(self, path, rel):
        full_path, content = self.try_path(path, rel)
        if content:
            return full_path, content
        raise RuntimeError('File not found')

    def render_jsonnet(self, manifeststr, tla_codes=None):
        try:
            json_str = _jsonnet.evaluate_snippet("snippet", manifeststr,
                                                 import_callback=self.import_callback,
                                                 native_callbacks=filters.jsonnet_callbacks(), tla_codes=tla_codes,
                                                 gc_min_objects=9999999, gc_growth_trigger=9999999)

        except RuntimeError as e:
            print "tla_codes: %s" % (str(tla_codes))
            print "\n".join(["%s %s" % (i, line) for i, line in
                             enumerate([l for l in manifeststr.split("\n") if re.match(r"^ *#", l) is None])])
            raise e
        return json.loads(json_str)



import json
import kpm
import kpm.registry
from kpm.commands.command_base import CommandBase


class VersionCmd(CommandBase):
    name = 'version'
    help_message = "show versions"

    def __init__(self, options):
        self.output = options.output
        self.api_version = None
        self.client_version = None
        self.registry_host = options.registry_host
        super(VersionCmd, self).__init__(options)

    @classmethod
    def _add_arguments(self, parser):
        parser.add_argument("-H", "--registry-host", nargs="?", default=kpm.registry.DEFAULT_REGISTRY,
                            help='registry API url')

    def _call(self):
        r = kpm.version(self.registry_host)
        self.api_version = r['api-version']
        self.client_version = r['client-version']

    def _render_json(self):
        print json.dumps({"api-version": self.api_version,
                          "client-version": self.client_version})

    def _render_console(self):
        print "Api-version: %s" % self.api_version
        print "Client-version: %s" % self.client_version



from flask import jsonify, Blueprint
from kpm.api.authorization import check_access
import kpm.api.builder
import kpm.api.registry

proxy_app = Blueprint('proxy', __name__,)


@proxy_app.route("/api/v1/packages/<path:package>/generate", methods=['GET'], strict_slashes=False)
@check_access('read')
def generate_package(package):
    return kpm.api.builder.build(package)


@proxy_app.route("/api/v1/packages/<path:package>/pull", methods=['GET'], strict_slashes=False)
@check_access('read')
def pull_package(package):
    return kpm.api.registry.pull(package)


@proxy_app.route("/api/v1/packages/count", methods=['GET'], strict_slashes=False)
def package_count(package):
    return jsonify({})


# @TODO filter privates
@proxy_app.route("/api/v1/packages", methods=['GET'], strict_slashes=False)
def packages():
    return kpm.api.registry.list_packages()


@proxy_app.route("/api/v1/packages/<path:package>", methods=['GET'], strict_slashes=False)
@check_access('read')
def package(package):
    return kpm.api.registry.show_package(package)


@proxy_app.route("/api/v1/packages/<path:package>", methods=['POST'], strict_slashes=False)
@proxy_app.route("/api/v1/packages", methods=['POST'], strict_slashes=False)
@check_access('write')
def push_package(package):
    return kpm.api.registry.push(package)


@proxy_app.route("/api/v1/packages/<path:package>/tree", methods=['GET'], strict_slashes=False)
@check_access('read')
def package_tree(package):
    return kpm.api.builder.tree(package)


@proxy_app.route("/api/v1/packages/<path:package>/file/<path:filepath>", methods=['GET'], strict_slashes=False)
@check_access('read')
def show_file(package, filepath):
    return kpm.api.builder.show_file(package, filepath)



import logging
import os.path
import yaml
from kpm.render_jsonnet import RenderJsonnet, yaml_to_jsonnet
from kpm.packager import authorized_files
from kpm.manifest import ManifestBase

__all__ = ['ManifestJsonnet']
logger = logging.getLogger(__name__)

MANIFEST_FILES = ["manifest.jsonnet", "manifest.yaml"]


class ManifestJsonnet(ManifestBase):
    def __init__(self, package=None, tla_codes=None):
        self.tla_codes = tla_codes
        if package is not None:
            self._load_from_package(package)
        else:
            self._load_from_path()

        super(ManifestJsonnet, self).__init__()

    def _load_from_package(self, package):
        if package.isjsonnet():
            self._load_jsonnet(package.manifest, package.files)
        else:
            self._load_yaml(package.manifest, package.files)

    def _load_from_path(self):
        for f in MANIFEST_FILES:
            if os.path.exists(f):
                mfile = f
                break
        _, ext = os.path.splitext(mfile)
        with open(mfile) as f:
            auth_files = authorized_files()
            files = dict(zip(auth_files, [None] * len(auth_files)))
            if ext == '.jsonnet':
                self._load_jsonnet(f.read(), files)
            else:
                self._load_yaml(f.read(), files)

    def _load_jsonnet(self, jsonnetstr, files):
        k = RenderJsonnet(files)
        r = k.render_jsonnet(jsonnetstr, self.tla_codes)
        self.update(r)

    def _load_yaml(self, yamlstr, files):
        try:
            jsonnetstr = yaml_to_jsonnet(yamlstr, self.tla_codes)
            files['manifest.jsonnet'] = jsonnetstr
            self._load_jsonnet(jsonnetstr, files)
        except yaml.YAMLError, exc:
            print "Error in configuration file:"
            if hasattr(exc, 'problem_mark'):
                mark = exc.problem_mark
                print "Error position: (%s:%s)" % (mark.line+1, mark.column+1)
            raise exc
